/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import static java.lang.Math.PI;
import static java.lang.Math.abs;
import static java.lang.Math.acos;
import static java.lang.Math.cos;
import static java.lang.Math.sin;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Hilfsklasse f&uuml;r allerlei geometrische Aufgaben. Der Schwerpunkt der
 * Hilfsmethoden liegt auf Berechnungen rund um Fuzzy.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */

public class Geometrie {

	/**
	 * Vergleicht zwei Punkte nur nach x-Koordinaten.
	 */
	public static class Point2DComparatorX implements Comparator<Point2D> {

		/**
		 * Vergleicht die x-Koordinaten der beiden Punkte.
		 *
		 * @param p1
		 *            Ein Punkt
		 * @param p2
		 *            Ein zweiter Punkt
		 * @return 0 bei Gleichheit, -1 wenn x1 kleiner, +1 wenn x1
		 *         gr&ouml;&szlig;er
		 * @see java.util.Comparator#compare(Object, Object)
		 */
		@Override
		public int compare(final Point2D p1, final Point2D p2) {
			int i = 0;

			if (p1.getX() < p2.getX()) {
				i = -1;
			} else if (p1.getX() > p2.getX()) {
				i = 1;
			}

			return i;
		}

	}

	/**
	 * Vergleicht zwei Punkte nur nach y-Koordinaten.
	 */
	public static class Point2DComparatorY implements Comparator<Point2D> {

		/**
		 * Vergleicht die y-Koordinaten der beiden Punkte.
		 *
		 * @param p1
		 *            Ein Punkt
		 * @param p2
		 *            Ein zweiter Punkt
		 * @return 0 bei Gleichheit, -1 wenn y1 kleiner, +1 wenn y1
		 *         gr&ouml;&szlig;er
		 * @see java.util.Comparator#compare(Object, Object)
		 */
		@Override
		public int compare(final Point2D p1, final Point2D p2) {
			int i = 0;

			if (p1.getY() < p2.getY()) {
				i = -1;
			} else if (p1.getY() > p2.getY()) {
				i = 1;
			}

			return i;
		}

	}

	/**
	 * Berechnet den Fl&auml;cheninhalt eines Dreiecks.
	 *
	 * @param p1
	 *            Erster Punkt des Dreiecks
	 * @param p2
	 *            Zweiter Punkt des Dreiecks
	 * @param p3
	 *            Dritter Punkt des Dreiecks
	 * @return Fl&auml;cheninhalt
	 */
	public static double flaeche(final Point2D p1, final Point2D p2,
			final Point2D p3) {
		double alpha; // Der Winkel am Punkt A
		final double a = p1.distance(p2); // Die Länge der Seite a
		final double b = p2.distance(p3); // Die Länge der Seite b
		final double c = p3.distance(p1); // Die Länge der Seite c
		double hc; // Die Länge der Höhe auf der Seite c
		Double flaeche;

		alpha = acos((((b * b) + (c * c)) - (a * a)) / (2 * b * c));
		hc = b * sin(alpha);

		flaeche = (c * hc) / 2;
		if (flaeche.equals(Double.NaN)) {
			return 0;
		}
		return flaeche;
	}

	/**
	 * Berechnet den Fl&auml;cheninhalt eines Rechtecks.
	 *
	 * @param r
	 *            Ein Rechteck
	 * @return Fl&auml;cheninhalt
	 */
	public static double flaeche(final Rectangle2D r) {
		return r.getWidth() * r.getHeight();
	}

	/**
	 * Berechnet die Fl&auml;che des Terms. Grundlage ist das Trapez des
	 * Fuzzy-Sets, welches als H&ouml;he die Zugeh&ouml;rigkeit annimmt.
	 *
	 * @param t
	 *            Ein Term
	 * @return Fl&auml;che des Terms mit der gegebenen Zugeh&ouml;rigkeit
	 */
	public static double flaeche(final Term t) {
		if (t.getZugehoerigkeit() == 0) {
			return 0;
		}

		return flaeche(trapezEcken(t));
	}

	/**
	 * Bestimmt einen Ortsvektor aus Fl&auml;che und Schwerpunkt. Der
	 * Wertebereich wird auf den Umfang eines Kreises abgebildet, d.h. auf 0 bis
	 * 360 bzw. 0 bis 2&pi; skaliert. Der Wert 0 (und 360 bzw. 2&pi;) entpricht
	 * als Vektor der x-Achse, der Wert 90 bzw. &pi;/2 entspricht als Vektor der
	 * y-Achse usw. Es wird ein Kartesisches Koordinatensystem angenommen.
	 *
	 * @param sx
	 *            X-Koordinate des Fl&auml;chenschwerpunkts
	 * @param a
	 *            Der Fl&auml;cheninhalt
	 * @param min
	 *            Das Minimum des Wertebereichs
	 * @param max
	 *            Das Maximum des Wertebereichs
	 *
	 * @return Ortsvektor als Punkt; die L&auml;nge entspricht der Fl&auml;che,
	 *         der Winkel zwischen der x-Achse und dem Vektor ist gleich dem
	 *         Betrag des Schwerpunkts des Wertebereichs auf 2&pi; skaliert
	 */
	public static Point2D ortsvektor(final double sx, final double a,
			final int min, final int max) {
		double x, y, s;

		// Schwerpunkt auf 2*PI skalieren
		s = (2 * PI * (sx - min)) / (max - min);

		x = a * cos(s);
		y = a * sin(s);

		return new Point2D.Double(x, y);
	}

	/**
	 * Berechnet den Schnittpunkt zweier Linien.
	 *
	 * @param linie1
	 *            Eine Linie
	 * @param linie2
	 *            Noch eine Linie
	 * @return Schnittpunkt beider Linien
	 */
	public static Point2D schnittpunkt(final Line2D linie1,
			final Line2D linie2) {
		if (!linie1.intersectsLine(linie2)) {
			return null;
		}

		// Verallgemeinerung der Strecken auf Geraden ist jetzt in Ordnung, da
		// wir bereits wissen, die beiden Strecken sich schneiden

		// Sonderfall: Beide Linien sind als Punkte entartet
		if ((Point2D.distance(linie1.getX1(), linie1.getX2(), linie1.getX2(),
				linie1.getY2()) > 0)
				&& (Point2D.distance(linie2.getX1(), linie2.getX2(),
						linie2.getX2(), linie2.getY2()) > 0)) {
			if ((linie1.getX1() == linie2.getX1())
					&& (linie1.getY1() == linie2.getY1())) {
				return new Point2D.Double(linie1.getX1(), linie1.getY1());
			}
		}

		// Allgemeine Geradengleichung: a * x + b * y + c = 0
		double a1, b1, c1; // Für erste Gerade
		double a2, b2, c2; // Für zweite Gerade
		double x, y;

		// Koeffizient der ersten Geradengleichung
		if (linie1.getX1() == linie1.getX2()) {
			// Senkrechte Gerade
			a1 = 1;
			b1 = 0;
			c1 = -linie1.getX1();
		} else if (linie1.getY1() == linie1.getY2()) {
			// Waagerechte Gerade
			a1 = 0;
			b1 = 1;
			c1 = -linie1.getY1();
		} else {
			// Schräge Gerade
			a1 = (linie1.getY2() - linie1.getY1())
					/ (linie1.getX2() - linie1.getX1());
			b1 = -1;
			c1 = linie1.getY1() - (a1 * linie1.getX1());
		}

		// Koeffizient der zweiten Geradengleichung
		if (linie2.getX1() == linie2.getX2()) {
			// Senkrechte Gerade
			a2 = 1;
			b2 = 0;
			c2 = -linie2.getX1();
		} else if (linie2.getY1() == linie2.getY2()) {
			// Waagerechte Gerade
			a2 = 0;
			b2 = 1;
			c2 = -linie2.getY1();
		} else {
			// Schräge Gerade
			a2 = (linie2.getY2() - linie2.getY1())
					/ (linie2.getX2() - linie2.getX1());
			b2 = -1;
			c2 = linie2.getY1() - (a2 * linie2.getX1());
		}

		// Schnittpunkt berechnen
		x = ((c2 * b1) - (c1 * b2)) / ((a1 * b2) - (a2 * b1));
		y = ((a2 * c1) - (a1 * c2)) / ((a1 * b2) - (a2 * b1));

		return new Point2D.Double(x, y);
	}

	/**
	 * Berechnet die x-Koordinate des Dreieckschwerpunkts.
	 *
	 * @param p1
	 *            Erster Punkt des Dreiecks
	 * @param p2
	 *            Zweiter Punkt des Dreiecks
	 * @param p3
	 *            Dritter Punkt des Dreiecks
	 * @return x-Koordinate des Schwerpunkts
	 */
	public static Point2D schwerpunkt(final Point2D p1, final Point2D p2,
			final Point2D p3) {
		final double x = (p1.getX() + p2.getX() + p3.getX()) / 3;
		final double y = (p1.getY() + p2.getY() + p3.getY()) / 3;

		return new Point2D.Double(x, y);
	}

	/**
	 * Berechnet die x-Koordinate des Rechteckschwerpunkts.
	 *
	 * @param r
	 *            Ein Rechteck
	 * @return x-Koordinate des Schwerpunkts
	 */
	public static Point2D schwerpunkt(final Rectangle2D r) {
		final Rectangle2D b = r.getBounds2D();

		final double x = b.getX() + (b.getWidth() / 2);
		final double y = b.getY() + (b.getHeight() / 2);

		return new Point2D.Double(x, y);
	}

	/**
	 * Berechnet den Schwerpunkt des Terms. Der Schwerpunkt basiert auf dem
	 * Trapez des Fuzzy-Sets, dessen H&ouml;he durch die Zugeh&ouml;rigkeit
	 * begrenzt wird.
	 *
	 * @param t
	 *            Ein Term
	 * @return Schwerpunkt des Terms bei gegebener Zugeh&ouml;rigkeit
	 */
	public static Point2D schwerpunkt(final Term t) {
		if (t.getZugehoerigkeit() == 0) {
			return null;
		}

		return schwerpunkt(trapezEcken(t));
	}

	/**
	 * Berechnet die L&auml;nge der Strecke ab mit Hilfe des Strahlensatz.
	 *
	 * <p>
	 * <img src="doc-files/strahlensatz.png" alt="Strahlensatz">
	 * </p>
	 *
	 * @param cd
	 *            Die Strecke cd
	 * @param za
	 *            Die Strecke za
	 * @param zc
	 *            Die Strecke zc
	 * @return L&auml;nge der Strecke ab
	 * @throws IllegalArgumentException
	 *             Wenn eine der Linien die L&auml;nge 0 hat
	 */
	public static double strahlensatzAB(final Line2D cd, final Line2D za,
			final Line2D zc) {
		final double laengeCD = cd.getP1().distance(cd.getP2());
		final double laengeZA = za.getP1().distance(za.getP2());
		final double laengeZC = zc.getP1().distance(zc.getP2());

		if ((laengeCD == 0) || (laengeZA == 0) || (laengeZC == 0)) {
			throw new IllegalArgumentException();
		}

		final double laengeAB = (laengeZA / laengeZC) * laengeCD;

		return laengeAB;
	}

	/**
	 * Berechnet die L&auml;nge der Strecke cd mit Hilfe des Strahlensatz.
	 *
	 * <p>
	 * <img src="doc-files/strahlensatz.png" alt="Strahlensatz">
	 * </p>
	 *
	 * @param ab
	 *            Die Strecke ab
	 * @param za
	 *            Die Strecke za
	 * @param zc
	 *            Die Strecke zc
	 * @return L&auml;nge der Strecke cd
	 */
	public static double strahlensatzCD(final Line2D ab, final Line2D za,
			final Line2D zc) {
		final double laengeAB = ab.getP1().distance(ab.getP2());
		final double laengeZA = za.getP1().distance(za.getP2());
		final double laengeZC = zc.getP1().distance(zc.getP2());

		if ((laengeAB == 0) || (laengeZA == 0) || (laengeZC == 0)) {
			throw new IllegalArgumentException();
		}

		final double laengeCD = (laengeAB / laengeZA) * laengeZC;

		return laengeCD;
	}

	/**
	 * Berechnet die L&auml;nge der Strecke za mit Hilfe des Strahlensatz.
	 *
	 * <p>
	 * <img src="doc-files/strahlensatz.png" alt="Strahlensatz">
	 * </p>
	 *
	 * @param ab
	 *            Die Strecke ab
	 * @param cd
	 *            Die Strecke cd
	 * @param zc
	 *            Die Strecke zc
	 * @return L&auml;nge der Strecke za
	 */
	public static double strahlensatzZA(final Line2D ab, final Line2D cd,
			final Line2D zc) {
		final double laengeAB = ab.getP1().distance(ab.getP2());
		final double laengeCD = cd.getP1().distance(cd.getP2());
		final double laengeZC = zc.getP1().distance(zc.getP2());

		if ((laengeAB == 0) || (laengeCD == 0) || (laengeZC == 0)) {
			throw new IllegalArgumentException();
		}

		final double laengeZA = (laengeAB / laengeCD) * laengeZC;

		return laengeZA;
	}

	/**
	 * Berechnet die L&auml;nge der Strecke zc mit Hilfe des Strahlensatz.
	 *
	 * <p>
	 * <img src="doc-files/strahlensatz.png" alt="Strahlensatz">
	 * </p>
	 *
	 * @param ab
	 *            Die Strecke ab
	 * @param cd
	 *            Die Strecke cd
	 * @param za
	 *            Die Strecke za
	 * @return L&auml;nge der Strecke zc
	 */
	public static double strahlensatzZC(final Line2D ab, final Line2D cd,
			final Line2D za) {
		final double laengeAB = ab.getP1().distance(ab.getP2());
		final double laengeCD = cd.getP1().distance(cd.getP2());
		final double laengeZA = za.getP1().distance(za.getP2());

		if ((laengeAB == 0) || (laengeCD == 0) || (laengeZA == 0)) {
			throw new IllegalArgumentException();
		}

		final double laengeZC = (laengeCD / laengeAB) * laengeZA;

		return laengeZC;
	}

	/**
	 * Bestimmt die Eckpunkte der Trapezkanten des Fuzzy-Sets. Die H&ouml;he des
	 * Trapez ist von der Zugeh&ouml;rigkeit (Trapezh&ouml;he) abh&auml;ngig.
	 *
	 * @param t
	 *            Ein Term (beinhaltet das Fuzzy-Set und die Zugeh&ouml;rigkeit)
	 * @return Liste der Trapezecken, nach x-Koordinaten sortiert
	 */
	public static List<Point2D> trapezEcken(final Term t) {
		final SortedSet<Point2D> punkte = new TreeSet<>(
				new Point2DComparatorX());

		for (final Line2D linie : trapezKanten(t)) {
			punkte.add(new Point2D.Double(linie.getX1(), linie.getY1()));
			punkte.add(new Point2D.Double(linie.getX2(), linie.getY2()));
		}

		return new ArrayList<>(punkte);
	}

	/**
	 * Bestimmt die drei Kanten die das Trapez des Fuzzy-Sets definieren. Die
	 * H&ouml;he des Trapez ist von der Zugeh&ouml;rigkeit (Trapezh&ouml;he)
	 * abh&auml;ngig. Die Grundseite des Trapez geh&ouml;rt nicht zu den Kanten!
	 *
	 * @param t
	 *            Ein Term (beinhaltet das Fuzzy-Set und die Zugeh&ouml;rigkeit)
	 * @return Unsortierte Liste der Kanten; drei f&uuml;r normale Fuzzy-Sets,
	 *         vier f&uuml;r den Spezialfall
	 * @throws IllegalArgumentException
	 *             Wenn die Zugh&ouml;rigkeit des Terms 0 ist
	 */
	public static Line2D[] trapezKanten(final Term t) {
		final double z = t.getZugehoerigkeit();

		if (t.getZugehoerigkeit() == 0) {
			throw new IllegalArgumentException();
		}

		final FuzzySet fs = t.getFuzzyVariable().getLinguistischeVariable()
				.getFuzzySet(t.getName());

		Line2D[] trapez;
		double x1, y1, x2, y2;

		if (fs.isSpezialfall()) {
			// 4: absteigende Rampe (2) + aufsteigende Rampe (2)
			trapez = new Line2D.Double[4];
		} else {
			// 3: aufsteigende Kante + horizontale Kante + absteigende Kante
			trapez = new Line2D.Double[3];
		}

		// Kante t1-t2
		x1 = fs.getT1();
		y1 = 0;
		x2 = strahlensatzX(fs.getT1(), fs.getT2(), z);
		y2 = z;
		trapez[0] = new Line2D.Double(x1, y1, x2, y2);

		// Kante t3-t4
		x1 = strahlensatzX(fs.getT4(), fs.getT3(), z);
		y1 = z;
		x2 = fs.getT4();
		y2 = 0;
		trapez[1] = new Line2D.Double(x1, y1, x2, y2);

		if (fs.isSpezialfall()) {
			// horizontale Kante der absteigenden Rampe
			x1 = t.getFuzzyVariable().getLinguistischeVariable().getMinimum();
			y1 = z;
			x2 = strahlensatzX(fs.getT4(), fs.getT3(), z);
			y2 = z;
			trapez[2] = new Line2D.Double(x1, y1, x2, y2);

			// horizontale Kante der aufsteigenden Rampe
			x1 = strahlensatzX(fs.getT1(), fs.getT2(), z);
			y1 = z;
			x2 = t.getFuzzyVariable().getLinguistischeVariable().getMaximum();
			y2 = z;
			trapez[3] = new Line2D.Double(x1, y1, x2, y2);
		} else {
			// Kante t2-t2
			x1 = strahlensatzX(fs.getT1(), fs.getT2(), z);
			y1 = z;
			x2 = strahlensatzX(fs.getT4(), fs.getT3(), z);
			y2 = z;
			trapez[2] = new Line2D.Double(x1, y1, x2, y2);
		}

		return trapez;
	}

	/**
	 * Bestimmt die Eckpunkte der Trapezkanten des Fuzzy-Sets. Die H&ouml;he des
	 * Trapez ist von der Zugeh&ouml;rigkeit (Trapezh&ouml;he) abh&auml;ngig.
	 *
	 * @param t
	 *            Ein Term (beinhaltet das Fuzzy-Set und die Zugeh&ouml;rigkeit)
	 * @return Aufsteigend sortierte Liste der Trapezecken (nur x-Koordinaten)
	 */
	public static List<Double> trapezStellen(final Term t) {
		final SortedSet<Double> stellen = new TreeSet<>();

		for (final Line2D linie : trapezKanten(t)) {
			stellen.add(linie.getX1());
			stellen.add(linie.getX2());
		}

		return new ArrayList<>(stellen);
	}

	/**
	 * F&uuml;hrt die Vektoraddition &uuml;ber alle angegebenen Vektoren aus.
	 *
	 * @param vektoren
	 *            Liste von Vektoren
	 * @return Ergebnis der Vektoraddition
	 */
	public static Point2D vektorAddition(final Point2D[] vektoren) {
		final Point2D s = new Point2D.Double();

		for (final Point2D p : vektoren) {
			s.setLocation(s.getX() + p.getX(), s.getY() + p.getY());
		}

		return s;
	}

	/**
	 * Berechnet mit die Zugeh&ouml;rigkeit zu einem Fuzzy-Set bei einem
	 * vorgegebenen Messwert.
	 *
	 * @param fs
	 *            Ein Fuzzy-Set
	 * @param wert
	 *            Der Messwert
	 * @return Die Zugeh&ouml;rigkeit des Fuzzy-Sets bei dem Messwert
	 */
	public static float zugehoerigkeit(final FuzzySet fs, final int wert) {
		Double z = 0.0;

		if ((fs.getT1() < wert) && (wert < fs.getT2())) {
			// Wert liegt auf ansteigender Kante
			z = strahlensatzAB(new Line2D.Double(fs.getT2(), 0, fs.getT2(), 1),
					new Line2D.Double(fs.getT1(), 0, wert, 0),
					new Line2D.Double(fs.getT1(), 0, fs.getT2(), 0));
		} else if ((fs.getT2() <= wert) && (wert <= fs.getT3())) {
			// Wert liegt auf horizontaler Kante, kein Spezialfall
			z = 1.0;
		} else if ((fs.getT3() < wert) && (wert < fs.getT4())) {
			// Wert liegt auf absteigender Kante
			z = strahlensatzAB(new Line2D.Double(fs.getT3(), 0, fs.getT3(), 1),
					new Line2D.Double(fs.getT4(), 0, wert, 0),
					new Line2D.Double(fs.getT4(), 0, fs.getT3(), 0));
		} else if (fs.isSpezialfall()) {
			// Wert liegt außerhalb der Fuzzy-Set-Grenzen und ist Spezialfall
			if ((fs.getT4() <= wert) && (wert <= fs.getT1())) {
				// Wert liegt zwischen den beiden Rampen
				z = 0.0;
			} else {
				// Wert liegt auf der Rampe
				z = 1.0;
			}
		}

		return z.floatValue();
	}

	/**
	 * Berechnet die Fl&auml;che unter dem Kantenzug. Die Punkte des Kantenzugs
	 * werden vor der Berechnung nach x-Koordinaten sortiert.
	 *
	 * @param kantenzug
	 *            Eine Punktliste
	 * @return Fl&auml;cheninhalt
	 * @throws IllegalArgumentException
	 *             Wenn der Kantenzug weniger als zwei Punkte enth&auml;lt
	 */
	static double flaeche(final Collection<Point2D> kantenzug) {
		if (kantenzug.size() < 2) {
			throw new IllegalArgumentException();
		}

		double a = 0; // Gesamtfläche unter dem Kantenzug
		final Queue<Point2D> queue = new PriorityQueue<>(20,
				new Point2DComparatorX());
		queue.addAll(kantenzug);

		Point2D p1;
		Point2D p2 = queue.poll();
		while (!queue.isEmpty()) {
			p1 = p2;
			p2 = queue.poll();

			if (p1.getY() == p2.getY()) {
				// Unter der Kante p1-p2 ist ein Rechteck
				p1 = new Point2D.Double(p1.getX(), 0); // Auf Grundlinie setzen
				final Rectangle2D r = new Rectangle2D.Double();
				r.setFrameFromDiagonal(p1, p2);
				a += flaeche(r);
			} else {
				// Unter der Kante p1-p2 ist ein Trapez
				// Genauer: Ein Dreieck was auf einem Rechteck steht

				Point2D p3; // Dritter Dreieckspunkt
				if (p1.getY() > p2.getY()) {
					p3 = new Point2D.Double(p1.getX(), p2.getY());
				} else {
					p3 = new Point2D.Double(p2.getX(), p1.getY());
				}

				// Dreieck berechnen
				a += flaeche(p1, p2, p3);

				// Rechteck berechnen
				final Rectangle2D r = new Rectangle2D.Double();
				if (p1.getY() > p2.getY()) {
					p3 = new Point2D.Double(p3.getX(), 0); // Auf Grundlinie
					// setzen
					r.setFrameFromDiagonal(p3, p2);
				} else {
					p1 = new Point2D.Double(p1.getX(), 0); // Auf Grundlinie
					// setzen
					r.setFrameFromDiagonal(p1, p3);
				}
				a += flaeche(r);
			}
		}

		return a;
	}

	/**
	 * Berechnet den Schwerpunkt der Fl&auml;che unter dem Kantenzug. Die Punkte
	 * des Kantenzugs werden vor der Berechnung nach x-Koordinaten sortiert.
	 *
	 * @param kantenzug
	 *            Eine Liste von Punkte
	 * @return Der Schwerpunkt
	 * @throws IllegalArgumentException
	 *             Wenn der Kantenzug weniger als zwei Punkte enth&auml;lt
	 */
	static Point2D schwerpunkt(final Collection<Point2D> kantenzug) {
		if (kantenzug.size() < 2) {
			throw new IllegalArgumentException();
		}

		double asx = 0; // Teilfläche * Teilschwerpunkt, x-Komponente
		double asy = 0; // Teilfläche * Teilschwerpunkt, y-Komponente
		final Queue<Point2D> queue = new PriorityQueue<>(20,
				new Point2DComparatorX());
		queue.addAll(kantenzug);

		Point2D p1;
		Point2D p2 = queue.poll();
		while (!queue.isEmpty()) {
			p1 = p2;
			p2 = queue.poll();

			double a = 0; // (Teil)fläche des Abschnitts
			Point2D s; // Schwerpunkt des Abschnitts

			if (p1.getY() == p2.getY()) {
				// Unter der Kante p1-p2 ist ein Rechteck
				p1 = new Point2D.Double(p1.getX(), 0); // Auf Grundlinie setzen
				final Rectangle2D r = new Rectangle2D.Double();
				r.setFrameFromDiagonal(p1, p2);
				a = flaeche(r);
				s = schwerpunkt(r);
				asx += a * s.getX();
				asy += a * s.getY();
			} else {
				// Unter der Kante p1-p2 ist ein Trapez
				// Genauer: Ein Dreieck was auf einem Rechteck steht

				Point2D p3; // Dritter Dreieckspunkt
				if (p1.getY() > p2.getY()) {
					p3 = new Point2D.Double(p1.getX(), p2.getY());
				} else {
					p3 = new Point2D.Double(p2.getX(), p1.getY());
				}

				// Dreieck berechnen
				a = flaeche(p1, p2, p3);
				s = schwerpunkt(p1, p2, p3);
				asx += a * s.getX();
				asy += a * s.getY();

				// Rechteck berechnen
				final Rectangle2D r = new Rectangle2D.Double();
				if (p1.getY() > p2.getY()) {
					p3 = new Point2D.Double(p3.getX(), 0); // Auf Grundlinie
					// setzen
					r.setFrameFromDiagonal(p3, p2);
				} else {
					p1 = new Point2D.Double(p1.getX(), 0); // Auf Grundlinie
					// setzen
					r.setFrameFromDiagonal(p1, p3);
				}
				a = flaeche(r);
				s = schwerpunkt(r);
				asx += a * s.getX();
				asy += a * s.getY();
			}
		}

		final double a = flaeche(kantenzug);
		if (a == 0) {
			return null;
		}

		return new Point2D.Double(asx / a, asy / a);
	}

	/**
	 * Berechnet mit Hilfe des Strahlensatz den Schnittpunkt bei einer
	 * bestimmten Zugeh&ouml;rigkeit.
	 * <p>
	 * a &lt;&gt; z. Die L&auml;nge der gegen&uuml;berliegenden Seite von z ist
	 * immer gleich 1. Die H&ouml;he steht senkrecht auf der Strecke az. Die
	 * Strecke az liegt auf der x-Achse.
	 * </p>
	 * <p>
	 * <img src="doc-files/strahlensatzX.png" alt="Strahlensatz">
	 * </p>
	 *
	 * @param z
	 *            Das Zentrum
	 * @param a
	 *            Ecke mit rechten Winkel
	 * @param y
	 *            H&ouml;he bzw. Zugeh&ouml;rigkeit
	 * @return x-Koordinate der H&ouml;he <code>y</code>
	 */
	private static double strahlensatzX(final double z, final double a,
			final double y) {
		if (z < a) {
			return z + (y * abs(z - a));
		}

		return z - (y * abs(z - a));
	}

	/**
	 * Konstruktor verstecken.
	 */
	protected Geometrie() {
		// nix
	}

}
