/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib.aggregation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyLibMsg;
import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.LinguistischeVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.Term;
import de.bsvrz.iav.fuzzylib.interpreter.logik.FuzzyLogikHandler;
import de.bsvrz.sys.funclib.bitctrl.interpreter.HandlerValidation;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterException;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterMessages;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Operator;

/**
 * Handelt die Aggregation von Fuzzy-Variablen. Die fuzzy-logischen
 * Verkn&uuml;pfungen werden analog auf alle gleichnamigen Terme der beteiligten
 * Fuzzy-Variablen angewandt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class AggregationHandler extends FuzzyLogikHandler {

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Object perform(final Operator operator,
			final List<Object> operanden) {
		if ((operator == null)
				|| !validiereHandler(operator, operanden).isValid()) {
			throw new InterpreterException(
					InterpreterMessages.HandlerNotFound.toString());
		}

		if (super.validiereHandler(operator, operanden).isValid()) {
			return super.perform(operator, operanden);
		}

		FuzzyVariable fv;

		// Ergebnis-Fuzzy-Variable konstruieren
		fv = ((FuzzyVariable) (operanden.get(0)));
		if (fv.getLinguistischeVariable() != null) {
			fv = fv.getLinguistischeVariable().fuzzyfiziere();
		} else {
			String[] terme;

			terme = fv.getTermnamen().toArray(new String[0]);
			fv = new FuzzyVariable(fv.getName(), terme);
		}

		final Map<String, Object[]> terme = baueTermlisten(operanden);

		// Operation analog für alle Terme der Fuzzy-Variable ausführen
		if ((operator == UND) || (operator == MINIMUM)) {
			for (final Term term : fv.getTerme()) {
				fv.setTerm(term.getName(),
						minimum(terme.get(term.getName())).getZugehoerigkeit());
			}
		} else if ((operator == ODER) || (operator == MAXIMUM)) {
			for (final Term term : fv.getTerme()) {
				fv.setTerm(term.getName(),
						maximum(terme.get(term.getName())).getZugehoerigkeit());
			}
		} else if (operator == IMPLIKATION) {
			for (final Term term : fv.getTerme()) {
				fv.setTerm(term.getName(),
						implikation(terme.get(term.getName()))
								.getZugehoerigkeit());
			}
		} else if (operator == ALGEBRAISCHES_PRODUKT) {
			for (final Term term : fv.getTerme()) {
				fv.setTerm(term.getName(),
						algebraischesProdukt(terme.get(term.getName()))
								.getZugehoerigkeit());
			}
		} else if (operator == ALGEBRAISCHE_SUMME) {
			for (final Term term : fv.getTerme()) {
				fv.setTerm(term.getName(),
						algebraischeSumme(terme.get(term.getName()))
								.getZugehoerigkeit());
			}
		} else if (operator == GAMMA_OPERATOR) {
			for (final Term term : fv.getTerme()) {
				fv.setTerm(term.getName(),
						gammaOperator(terme.get(term.getName()))
								.getZugehoerigkeit());
			}
		}

		return fv;
	}

	/**
	 * NICHT bzw KOMPLEMENT werden f&uuml;r Fuzzy-Variablen nicht
	 * unterst&uuml;tzt, da zum Aggregieren mindestens zwei Variablen notwendig
	 * sind.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public HandlerValidation validiereHandler(final Operator operator,
			final List<? extends Object> operanden) {
		if (operanden == null) {
			throw new NullPointerException(
					FuzzyLibMsg.OperandenListeIstNull.toString());
		}

		boolean ok = false;

		// Anzahl der Operanden und Operation prüfen
		switch (operanden.size()) {
		case 2:
			if (IMPLIKATION.equals(operator)
					|| ALGEBRAISCHE_SUMME.equals(operator)
					|| ALGEBRAISCHES_PRODUKT.equals(operator)) {
				ok = true;
			}
			break;
		case 3:
			if (GAMMA_OPERATOR.equals(operator)) {
				ok = true;
			}
			break;
		default:
			// nix
		}

		if (UND.equals(operator) || ODER.equals(operator)
				|| MAXIMUM.equals(operator) || MINIMUM.equals(operator)) {
			ok = true;
		}

		// Alle Operanden müssen eine FuzzyVariable mit der selben
		// linguistischen Variable sein
		LinguistischeVariable lv = null;
		for (final Object obj : operanden) {
			if (obj instanceof FuzzyVariable) {
				final FuzzyVariable fv = (FuzzyVariable) obj;

				if (lv == null) {
					lv = fv.getLinguistischeVariable();
				}

				if (!lv.equals(fv.getLinguistischeVariable())) {
					ok = false;
					break;
				}
			}
		}

		HandlerValidation result = null;
		if (ok) {
			result = new HandlerValidation(true, true);
		} else {
			result = super.validiereHandler(operator, operanden);
		}

		// Handler muss auch Aufgaben des abgeleiteten Handlers übernehmen
		return result;
	}

	/**
	 * Baut aus der Liste der Fuzzy-Variablen eine Liste von Termen. Dabei
	 * werden die gleichnamigen Terme aller Fuzzy-Variablen zusammengefasst.
	 *
	 * @param operanden
	 *            Liste von Fuzzy-Variablen
	 * @return Liste von Termen, nach Name sortiert
	 */
	protected static Map<String, Object[]> baueTermlisten(
			final List<? extends Object> operanden) {
		final Map<String, Object[]> terme = new HashMap<>();

		for (final Object obj : operanden) {
			final FuzzyVariable fv = (FuzzyVariable) obj;
			for (final Term term : fv.getTerme()) {
				if (terme.containsKey(term.getName())) {
					final List<Object> dingens = Arrays
							.asList(terme.get(term.getName()));
					final List<Object> liste = new ArrayList<>(dingens);
					liste.add(term);
					terme.put(term.getName(), liste.toArray());
				} else {
					terme.put(term.getName(), new Object[] { term });
				}
			}
		}

		return terme;
	}

}
